{-|
Module      : Gargantext.API.Node.Corpus.Export.Utils
Description : Corpus export
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX
-}

{-# LANGUAGE TypeApplications #-}

module Gargantext.API.Node.Corpus.Export.Utils
  where

import Control.Exception.Safe qualified as CES
import Data.Aeson qualified as Aeson
import Data.ByteString.Lazy qualified as BSL
import Data.HashMap.Strict qualified as HashMap
import Data.Map.Strict qualified as Map
import Data.Set qualified as Set
import Data.Version (showVersion)
import Database.SQLite.Simple qualified as S
import Gargantext.API.Ngrams.Tools (filterListWithRoot, mapTermListRoot, getRepo)
import Gargantext.API.Ngrams.Types ( NgramsTerm(unNgramsTerm) )
import Gargantext.API.Node.Corpus.Export.Types (CorpusSQLite(..))
import Gargantext.Core.NodeStory.Types ( HasNodeStoryEnv, NodeListStory )
import Gargantext.Core.Text.Ngrams (NgramsType(NgramsTerms))
import Gargantext.Core.Types (CorpusId, ListId, NodeType(NodeList))
import Gargantext.Core.Types.Main (ListType(..))
import Gargantext.Database.Action.Metrics.NgramsByContext (getNgramsByContextOnlyUser)
import Gargantext.Database.Admin.Config (userMaster)
import Gargantext.Database.Admin.Types.Hyperdata.Corpus (HyperdataCorpus)
import Gargantext.Database.Admin.Types.Hyperdata.List (HyperdataList)
import Gargantext.Database.Admin.Types.Node (unNodeId, ContextId(_ContextId))
import Gargantext.Database.Prelude (DBCmd, IsDBCmd)
import Gargantext.Database.Schema.Context (context_id, context_name, context_date, context_hyperdata) 
import Gargantext.Database.Schema.Node (node_hash_id, node_hyperdata, node_name, node_parent_id)
import Gargantext.Database.Query.Table.Node ( defaultList, getNodeWith )
import Gargantext.Database.Query.Table.NodeContext (selectDocNodes)
import Gargantext.Database.Query.Table.Node.Error (HasNodeError)
import Gargantext.Database.Query.Table.Node.Select (selectNodesWithUsername)
import Gargantext.Prelude
import Paths_gargantext qualified as PG -- cabal magic build module
import System.Directory (removeDirectoryRecursive)
import System.IO.Temp (createTempDirectory, getCanonicalTemporaryDirectory)




getContextNgrams :: HasNodeError err
                 => CorpusId
                 -> ListId
                 -> ListType
                 -> NgramsType
                 -> NodeListStory
                 -> DBCmd err (Map ContextId (Set NgramsTerm))
getContextNgrams cId lId listType nt repo = do
--  lId <- case lId' of
--    Nothing -> defaultList cId
--    Just  l -> pure l

  lIds <- selectNodesWithUsername NodeList userMaster
  let ngs = filterListWithRoot [listType] $ mapTermListRoot [lId] nt repo
  -- TODO HashMap
  r <- getNgramsByContextOnlyUser cId (lIds <> [lId]) nt (HashMap.keys ngs)
  pure r


mkCorpusSQLite :: ( CES.MonadMask m
                  , HasNodeStoryEnv env
                  , HasNodeError err
                  , IsDBCmd env err m )
               => CorpusId
               -> Maybe ListId
               -> m CorpusSQLite
mkCorpusSQLite cId lId =
  CES.bracket setup tearDown $ \(fp, _fname, fpath) -> do

  corpus <- getNodeWith cId (Proxy @HyperdataCorpus)
  
  listId <- case lId of
        Nothing -> defaultList cId
        Just l  -> pure l

  l <- getNodeWith listId (Proxy @HyperdataList)

  repo <- getRepo [listId]
  let nt = NgramsTerms
  mapNgs  <- getContextNgrams cId listId MapTerm nt repo
  stopNgs  <- getContextNgrams cId listId StopTerm nt repo
  candidateNgs  <- getContextNgrams cId listId CandidateTerm nt repo
  docs <- selectDocNodes cId
        
  liftBase $ putText $ "[mkCorpusSQLite] listId: " <> show listId
  liftBase $ putText $ "[mkCorpusSQLite] fp: " <> show fp

  liftBase $ S.withConnection fpath $ \conn -> do
    -- better performance
    -- https://kerkour.com/sqlite-for-servers
    S.execute_ conn "PRAGMA journal_mode = WAL"
    
    S.execute_ conn "CREATE TABLE info (key, value);"
    S.execute conn "INSERT INTO info (key, value) VALUES ('gargVersion', ?)" (S.Only $ showVersion PG.version)
    S.execute conn "INSERT INTO info (key, value) VALUES ('corpusId', ?)" (S.Only $ unNodeId cId)
    S.execute conn "INSERT INTO info (key, value) VALUES ('listId', ?)" (S.Only $ unNodeId listId)
    S.execute_ conn "INSERT INTO info (key, value) VALUES ('created', datetime())"

    S.execute_ conn "CREATE TABLE corpus (id, name, hash, parent_id, hyperdata)"
    S.execute conn "INSERT INTO corpus (id, name, hash, parent_id, hyperdata) VALUES (?, ?, ?, ?, ?)"
      (unNodeId cId, corpus ^. node_name, corpus ^. node_hash_id, unNodeId <$> (corpus ^. node_parent_id), Aeson.encode (corpus ^. node_hyperdata))

    S.execute_ conn "CREATE TABLE lists (id, name, parent_id, hyperdata)"
    S.execute conn "INSERT INTO lists (id, name, parent_id, hyperdata) VALUES (?, ?, ?, ?)"
      (unNodeId listId, l ^. node_name, unNodeId <$> (l ^. node_parent_id), Aeson.encode (l ^. node_hyperdata))

    S.execute_ conn "CREATE TABLE ngrams (context_id, terms, type_)"
    let insertTerms ngs type_ = do
                      let ngs' = concatMap (\(ctxId, ngrams) -> (\n -> (_ContextId ctxId, unNgramsTerm n, type_)) <$> Set.toList ngrams) (Map.toList ngs)
                      S.executeMany conn "INSERT INTO ngrams (context_id, terms, type_) VALUES (?, ?, ?)" ngs'
    insertTerms mapNgs ("map" :: Text)
    insertTerms stopNgs ("stop" :: Text)
    insertTerms candidateNgs ("candidate" :: Text)

    S.execute_ conn "CREATE TABLE documents (context_id, name, date, hyperdata)"
    S.executeMany conn "INSERT INTO documents (context_id, name, date, hyperdata) VALUES (?, ?, ?, ?)"
      ((\c -> (unNodeId (c ^. context_id), c ^. context_name, c ^. context_date, Aeson.encode (c ^. context_hyperdata))) <$> docs)

  bsl <- liftBase $ BSL.readFile fpath
  
  pure $ CorpusSQLite bsl

  where
    setup = do
      tmpDir <- liftBase getCanonicalTemporaryDirectory
      fp <- liftBase $ createTempDirectory tmpDir "sqlite" 
      let fname = "gargantext.sqlite"
      let fpath = fp <> "/" <> fname

      pure (fp, fname, fpath)
    tearDown (fp, _fname, _fpath) = do
      liftBase $ removeDirectoryRecursive fp
    
